@@ -40,9 +40,9 @@ void *simple_loop(void *arg1) {
 	long core_id = (long) arg1;
 
 	struct wsgi_request *wsgi_req = uwsgi.wsgi_requests[core_id];
-	int i;
 
 #ifdef UWSGI_THREADING
+	int i;
 	//PyThreadState *pts;
 	sigset_t smask;
 
@@ -55,8 +55,11 @@ void *simple_loop(void *arg1) {
 		if (core_id > 0) {
 			// block all signals on new threads
 			sigfillset(&smask);
+#ifdef UWSGI_DEBUG
+			sigdelset(&smask, SIGSEGV);
+#endif
 			pthread_sigmask(SIG_BLOCK, &smask, NULL);
-			for (i = 0; i < 0xFF; i++) {
+			for (i = 0; i < 256; i++) {
 				if (uwsgi.p[i]->init_thread) {
 					uwsgi.p[i]->init_thread(core_id);
 				}
@@ -72,6 +75,7 @@ void *simple_loop(void *arg1) {
 
 	if (uwsgi.signal_socket > -1) {
 		event_queue_add_fd_read(main_queue, uwsgi.signal_socket);
+		event_queue_add_fd_read(main_queue, uwsgi.my_signal_socket);
 	}
 
 	while (uwsgi.workers[uwsgi.mywid].manage_next_request) {
@@ -115,7 +119,7 @@ void *zeromq_loop(void *arg1) {
 
 	struct wsgi_request *wsgi_req = uwsgi.wsgi_requests[core_id];
 	uwsgi.zeromq_recv_flag = 0;
-	zmq_pollitem_t zmq_poll_items[2];
+	zmq_pollitem_t zmq_poll_items[3];
 	char uwsgi_signal;
 
 	if (uwsgi.threads > 1) {
@@ -127,8 +131,11 @@ void *zeromq_loop(void *arg1) {
 		if (core_id > 0) {
 			// block all signals on new threads
 			sigfillset(&smask);
+#ifdef UWSGI_DEBUG
+			sigdelset(&smask, SIGSEGV);
+#endif
 			pthread_sigmask(SIG_BLOCK, &smask, NULL);
-			for (i = 0; i < 0xFF; i++) {
+			for (i = 0; i < 256; i++) {
 				if (uwsgi.p[i]->init_thread) {
 					uwsgi.p[i]->init_thread(core_id);
 				}
@@ -159,6 +166,10 @@ void *zeromq_loop(void *arg1) {
 		zmq_poll_items[1].socket = NULL;
 		zmq_poll_items[1].fd = uwsgi.signal_socket;
 		zmq_poll_items[1].events = ZMQ_POLLIN;
+
+		zmq_poll_items[2].socket = NULL;
+		zmq_poll_items[2].fd = uwsgi.my_signal_socket;
+		zmq_poll_items[2].events = ZMQ_POLLIN;
 	}
 
 
@@ -171,7 +182,7 @@ void *zeromq_loop(void *arg1) {
 
 
 		if (uwsgi.signal_socket > -1) {
-			if (zmq_poll(zmq_poll_items, 2, -1) < 0) {
+			if (zmq_poll(zmq_poll_items, 3, -1) < 0) {
 				uwsgi_error("zmq_poll()");
 				continue;
 			}
@@ -184,7 +195,9 @@ void *zeromq_loop(void *arg1) {
                         		}
                 		}
                 		else {
+#ifdef UWSGI_DEBUG
                         		uwsgi_log_verbose("master sent signal %d to worker %d\n", uwsgi_signal, uwsgi.mywid);
+#endif
                         		if (uwsgi_signal_handler(uwsgi_signal)) {
                                 		uwsgi_log_verbose("error managing signal %d on worker %d\n", uwsgi_signal, uwsgi.mywid);
                         		}
@@ -192,6 +205,26 @@ void *zeromq_loop(void *arg1) {
 				continue;
 			}
 
+			if (zmq_poll_items[2].revents & ZMQ_POLLIN) {
+                                if (read(uwsgi.my_signal_socket, &uwsgi_signal, 1) <= 0) {
+                                        if (uwsgi.no_orphans) {
+                                                uwsgi_log_verbose("uWSGI worker %d screams: UAAAAAAH my master died, i will follow him...\n", uwsgi.mywid);
+                                                end_me(0);
+                                        }
+                                }
+                                else {
+#ifdef UWSGI_DEBUG
+                                        uwsgi_log_verbose("master sent signal %d to worker %d\n", uwsgi_signal, uwsgi.mywid);
+#endif
+                                        if (uwsgi_signal_handler(uwsgi_signal)) {
+                                                uwsgi_log_verbose("error managing signal %d on worker %d\n", uwsgi_signal, uwsgi.mywid);
+                                        }
+                                }
+                                continue;
+                        }
+
+			
+
 			if (zmq_poll_items[0].revents & ZMQ_POLLIN) {
 				wsgi_req->poll.fd = wsgi_req->socket->proto_accept(wsgi_req, uwsgi.zmq_socket->fd);
 			}
