@@ -9,11 +9,12 @@ extern struct uwsgi_perl uperl;
 struct uwsgi_perl uperl;
 #endif
 
-struct option uwsgi_perl_options[] = {
+struct uwsgi_option uwsgi_perl_options[] = {
 
-        {"psgi", required_argument, 0, LONG_ARGS_PSGI},
-        {"perl-local-lib", required_argument, 0, LONG_ARGS_PERL_LOCAL_LIB},
-        {0, 0, 0, 0},
+        {"psgi", required_argument, 0, "load a psgi app", uwsgi_opt_set_str, &uperl.psgi, 0},
+        {"perl-no-die-catch", no_argument, 0, "do not catch $SIG{__DIE__}", uwsgi_opt_true, &uperl.no_die_catch, 0},
+        {"perl-local-lib", required_argument, 0, "set perl locallib path", uwsgi_opt_set_str, &uperl.locallib, 0},
+        {0, 0, 0, 0, 0, 0, 0},
 
 };
 
@@ -22,8 +23,7 @@ extern struct http_status_codes hsc[];
 SV *uwsgi_perl_obj_new(char *class, size_t class_len) {
 
 	SV *newobj;
-	// set current context ?
-	//dTHX;
+
 	dSP;
 
 	ENTER;
@@ -48,21 +48,21 @@ SV *uwsgi_perl_obj_new(char *class, size_t class_len) {
 SV *uwsgi_perl_call_stream(SV *func) {
 
 	SV *ret = NULL;
-        // set current context ?
-        //dTHX;
-        dSP;
+	struct wsgi_request *wsgi_req = current_wsgi_req();
+	struct uwsgi_app *wi = &uwsgi_apps[wsgi_req->app_id];
 
+        dSP;
         ENTER;
         SAVETMPS;
         PUSHMARK(SP);
-        XPUSHs( sv_2mortal(newRV((SV*) uperl.stream_responder)));
+        XPUSHs( sv_2mortal(newRV((SV*) ((SV **)wi->responder0)[wsgi_req->async_id])));
         PUTBACK;
 
 	call_sv( func, G_SCALAR | G_EVAL);
 
 	SPAGAIN;
         if(SvTRUE(ERRSV)) {
-                uwsgi_log("%s\n", SvPV_nolen(ERRSV));
+                uwsgi_log("[uwsgi-perl error] %s\n", SvPV_nolen(ERRSV));
         }
         else {
                 ret = SvREFCNT_inc(POPs);
@@ -78,8 +78,7 @@ SV *uwsgi_perl_call_stream(SV *func) {
 int uwsgi_perl_obj_can(SV *obj, char *method, size_t len) {
 
 	int ret;
-        // set current context ? needed for threading
-        //dTHX;
+
         dSP;
 
         ENTER;
@@ -102,7 +101,6 @@ int uwsgi_perl_obj_can(SV *obj, char *method, size_t len) {
 
 }
 
-
 SV *uwsgi_perl_obj_call(SV *obj, char *method) {
 
         SV *ret = NULL;
@@ -149,11 +147,12 @@ AV *psgi_call(struct wsgi_request *wsgi_req, SV *psgi_func, SV *env) {
 	PUTBACK;
 
 	call_sv(psgi_func, G_SCALAR | G_EVAL);
+
 	SPAGAIN;
 
         if(SvTRUE(ERRSV)) {
                 internal_server_error(wsgi_req, "exception raised");
-                uwsgi_log("%s\n", SvPV_nolen(ERRSV));
+                uwsgi_log("[uwsgi-perl error] %s\n", SvPV_nolen(ERRSV));
         }
 	else {
 		ret = (AV *) SvREFCNT_inc(SvRV(POPs));
@@ -280,11 +279,20 @@ int uwsgi_perl_init(){
 
 	struct http_status_codes *http_sc;
 
-	int argc = 3;
+	int argc;
+	int i;
+
 	uperl.embedding[0] = "";
 	uperl.embedding[1] = "-e";
 	uperl.embedding[2] = "0";
 
+#ifndef USE_ITHREADS
+	if (uwsgi.threads > 1) {
+		uwsgi_log("your Perl environment does not support threads\n");
+		exit(1);
+	} 
+#endif
+
 	if (setenv("PLACK_ENV", "uwsgi", 0)) {
 		uwsgi_error("setenv()");
 	}
@@ -293,128 +301,102 @@ int uwsgi_perl_init(){
 		uwsgi_error("setenv()");
 	}
 
-#ifdef PERL_VERSION_STRING
-	uwsgi_log("initializing Perl %s environment\n", PERL_VERSION_STRING);
-#else
-	uwsgi_log("initializing Perl environment\n");
-#endif
+	argc = 3;
+
 	PERL_SYS_INIT3(&argc, (char ***) &uperl.embedding, &environ);
-	uperl.main = perl_alloc();
-	if (!uperl.main) {
-		uwsgi_log("unable to allocate perl interpreter\n");
+
+	uperl.main = uwsgi_calloc(sizeof(PerlInterpreter *) * uwsgi.threads);
+
+	uperl.main[0] = uwsgi_perl_new_interpreter();
+	if (!uperl.main[0]) {
 		return -1;
 	}
 
-	dTHXa(uperl.main);
-	PERL_SET_CONTEXT(uperl.main);
+	for(i=1;i<uwsgi.threads;i++) {
+		uperl.main[i] = uwsgi_perl_new_interpreter();
+                if (!uperl.main[i]) {
+                	uwsgi_log("unable to create new perl interpreter for thread %d\n", i+1);
+                        exit(1);
+                }
+	}
 
-	PL_perl_destruct_level = 2;
-	PL_origalen = 1;
-	perl_construct(uperl.main);
+	PERL_SET_CONTEXT(uperl.main[0]);
 
 	// filling http status codes
 	for (http_sc = hsc; http_sc->message != NULL; http_sc++) {
 		http_sc->message_size = strlen(http_sc->message);
 	}
 
-	PL_origalen = 1;
+#ifdef PERL_VERSION_STRING
+	uwsgi_log("initialized Perl %s main interpreter at %p\n", PERL_VERSION_STRING, uperl.main[0]);
+#else
+	uwsgi_log("initialized Perl main interpreter at %p\n", uperl.main[0]);
+#endif
 
 	return 1;
 
 }
 
-#ifdef my_perl
-void uwsgi_perl_enable_threads() {
-	
-	int i;
-
-	if (pthread_key_create(&uperl.u_interpreter, NULL)) {
-        	uwsgi_error("pthread_key_create()");
-                exit(1);
-	}
-
-	uperl.interp = malloc( sizeof(PerlInterpreter*) * uwsgi.threads );
-	if (!uperl.interp) {
-		uwsgi_error("malloc()");
-		exit(1);
-	}
-
-	for(i=1;i<uwsgi.threads;i++) {
-		uperl.interp[i] = perl_clone(uperl.main, CLONEf_KEEP_PTR_TABLE);
-		if (!uperl.interp[i]) {
-			uwsgi_log("unable to create new perl interpreter\n");
-			exit(1);
-		}
-	}
-
-	uperl.psgi_func = malloc( sizeof(SV*) * uwsgi.threads );
-	if (!uperl.psgi_func) {
-		uwsgi_error("malloc()");
-		exit(1);
-	}
-	
-
-	dTHXa(uperl.main);
-	PERL_SET_CONTEXT(uperl.main);
-
-	uperl.psgi_main = perl_eval_pv(uperl.psgibuffer, 0);
-	if (!uperl.psgi_main) {
-		uwsgi_log("unable to find PSGI function entry point.\n");
-		exit(1);
-	}
-
-	if(SvTRUE(ERRSV)) {
-		uwsgi_log("%s\n", SvPV_nolen(ERRSV));
-		exit(1);
-	}
-	
-
-}
-#endif
-
-
-
 int uwsgi_perl_request(struct wsgi_request *wsgi_req) {
 
 	SV **harakiri;
-	SV *psgi_func = uperl.psgi_main;
-	// ugly hack
-	register PerlInterpreter *my_perl = uperl.main;
 
 #ifdef UWSGI_ASYNC
 	if (wsgi_req->async_status == UWSGI_AGAIN) {
-		return psgi_response(wsgi_req, my_perl, wsgi_req->async_placeholder);	
+		return psgi_response(wsgi_req, wsgi_req->async_placeholder);	
 	}
 #endif
-
 	/* Standard PSGI request */
 	if (!wsgi_req->uh.pktsize) {
 		uwsgi_log("Invalid PSGI request. skip.\n");
 		return -1;
 	}
 
-
 	if (uwsgi_parse_vars(wsgi_req)) {
 		return -1;
 	}
 
+	wsgi_req->app_id = uwsgi_get_app_id(wsgi_req->appid, wsgi_req->appid_len, 5);
+	// if it is -1, try to load a dynamic app
+	if (wsgi_req->app_id == -1) {
+		if (wsgi_req->dynamic) {
+			if (uwsgi.threads > 1) {
+                                pthread_mutex_lock(&uperl.lock_loader);
+                        }
 
-#ifdef UWSGI_THREADING
-	if (uwsgi.threads > 1 && wsgi_req->async_id > 0) {
-		psgi_func = uperl.psgi_func[wsgi_req->async_id];
-		my_perl = pthread_getspecific(uperl.u_interpreter);
+			if (wsgi_req->script_len > 0) {
+				wsgi_req->app_id = init_psgi_app(wsgi_req, wsgi_req->script, wsgi_req->script_len, NULL);	
+			}
+			else if (wsgi_req->file_len > 0) {
+				wsgi_req->app_id = init_psgi_app(wsgi_req, wsgi_req->file, wsgi_req->file_len, NULL);	
+			}
+
+			if (uwsgi.threads > 1) {
+                                pthread_mutex_unlock(&uperl.lock_loader);
+                        }
+		}
+
+		if (wsgi_req->app_id == -1) {
+			internal_server_error(wsgi_req, "Perl application not found");	
+			// nothing to clear/free
+			return UWSGI_OK;
+		}
 	}
-#endif
 
+	struct uwsgi_app *wi = &uwsgi_apps[wsgi_req->app_id];
+	wi->requests++;
+
+	if (((PerlInterpreter **)wi->interpreter)[wsgi_req->async_id] != uperl.main[wsgi_req->async_id]) {
+		PERL_SET_CONTEXT(((PerlInterpreter **)wi->interpreter)[wsgi_req->async_id]);
+	}
 
 	ENTER;
 	SAVETMPS;
 
-
 	wsgi_req->async_environ = build_psgi_env(wsgi_req);
 	if (!wsgi_req->async_environ) goto clear;
 
-	wsgi_req->async_result = psgi_call(wsgi_req, psgi_func, wsgi_req->async_environ);
+	wsgi_req->async_result = psgi_call(wsgi_req, ((SV **)wi->callable)[wsgi_req->async_id], wsgi_req->async_environ);
 	if (!wsgi_req->async_result) goto clear;
 
 	if (SvTYPE((AV *)wsgi_req->async_result) == SVt_PVCV) {
@@ -428,7 +410,7 @@ int uwsgi_perl_request(struct wsgi_request *wsgi_req) {
 		goto clear2;
 	}
 
-	while (psgi_response(wsgi_req, my_perl, wsgi_req->async_result) != UWSGI_OK) {
+	while (psgi_response(wsgi_req, wsgi_req->async_result) != UWSGI_OK) {
 #ifdef UWSGI_ASYNC
 		if (uwsgi.async > 1) {
 			FREETMPS;
@@ -452,6 +434,11 @@ clear:
 	FREETMPS;
 	LEAVE;
 
+	// restore main interpreter if needed
+	if (((PerlInterpreter **)wi->interpreter)[wsgi_req->async_id] != uperl.main[wsgi_req->async_id]) {
+		PERL_SET_CONTEXT(uperl.main[wsgi_req->async_id]);
+	}
+
 	return UWSGI_OK;
 }
 
@@ -468,41 +455,6 @@ void uwsgi_perl_after_request(struct wsgi_request *wsgi_req) {
 
 }
 
-void uwsgi_perl_init_thread(int core_id) {
-
-
-	pthread_setspecific(uperl.u_interpreter, uperl.interp[core_id]);
-	dTHXa(uperl.interp[core_id]);
-	PERL_SET_CONTEXT(uperl.interp[core_id]);
-
-	uperl.psgi_func[core_id] = perl_eval_pv(uperl.psgibuffer, 0);
-        if (!uperl.psgi_func[core_id]) {
-        	uwsgi_log("unable to find PSGI function entry point.\n");
-		exit(1);
-        }
-
-        if(SvTRUE(ERRSV)) {
-        	uwsgi_log("%s\n", SvPV_nolen(ERRSV));
-		exit(1);
-        }
-
-
-}
-
-int uwsgi_perl_manage_options(int i, char *optarg) {
-
-        switch(i) {
-                case LONG_ARGS_PSGI:
-                        uperl.psgi = optarg;
-                        return 1;
-                case LONG_ARGS_PERL_LOCAL_LIB:
-                        uperl.locallib = optarg;
-                        return 1;
-        }
-
-        return 0;
-}
-
 int uwsgi_perl_magic(char *mountpoint, char *lazy) {
 
         if (!strcmp(lazy+strlen(lazy)-5, ".psgi")) {
@@ -518,6 +470,56 @@ int uwsgi_perl_magic(char *mountpoint, char *lazy) {
 
 }
 
+// taken from Torsten Foertsch AfterFork.xs
+void uwsgi_perl_post_fork() {
+
+	GV *tmpgv = gv_fetchpv("$", TRUE, SVt_PV);
+        if (tmpgv) {
+                SvREADONLY_off(GvSV(tmpgv));
+                sv_setiv(GvSV(tmpgv), (IV)getpid());
+                SvREADONLY_on(GvSV(tmpgv));
+        }
+
+}
+
+int uwsgi_perl_mount_app(char *mountpoint, char *app, int regexp) {
+
+	if (uwsgi_endswith(app, ".pl") || uwsgi_endswith(app, ".psgi")) {
+        	uwsgi.wsgi_req->appid = mountpoint;
+        	uwsgi.wsgi_req->appid_len = strlen(mountpoint);
+
+        	return init_psgi_app(uwsgi.wsgi_req, app, strlen(app), NULL);
+	}
+	return -1;
+
+}
+
+void uwsgi_perl_init_thread(int core_id) {
+
+#ifdef USE_ITHREADS
+        PERL_SET_CONTEXT(uperl.main[core_id]);
+#endif
+}
+
+void uwsgi_perl_pthread_prepare(void) {
+        pthread_mutex_lock(&uperl.lock_loader);
+}
+
+void uwsgi_perl_pthread_parent(void) {
+        pthread_mutex_unlock(&uperl.lock_loader);
+}
+
+void uwsgi_perl_pthread_child(void) {
+        pthread_mutex_init(&uperl.lock_loader, NULL);
+}
+
+
+void uwsgi_perl_enable_threads(void) {
+#ifdef USE_ITHREADS
+	pthread_mutex_init(&uperl.lock_loader, NULL);
+	pthread_atfork(uwsgi_perl_pthread_prepare, uwsgi_perl_pthread_parent, uwsgi_perl_pthread_child);
+#endif
+}
 
 struct uwsgi_plugin psgi_plugin = {
 
@@ -525,16 +527,16 @@ struct uwsgi_plugin psgi_plugin = {
 	.modifier1 = 5,
 	.init = uwsgi_perl_init,
 	.options = uwsgi_perl_options,
+
 	.init_apps = uwsgi_psgi_app,
-	//.magic = uwsgi_perl_magic,
-	//.help = uwsgi_perl_help,
-#ifdef my_perl
-	.enable_threads = uwsgi_perl_enable_threads,
-#endif
-	.manage_opt = uwsgi_perl_manage_options,
+	.mount_app = uwsgi_perl_mount_app,
+
 	.init_thread = uwsgi_perl_init_thread,
+
+	.post_fork = uwsgi_perl_post_fork,
 	.request = uwsgi_perl_request,
 	.after_request = uwsgi_perl_after_request,
+	.enable_threads = uwsgi_perl_enable_threads,
 
 	.magic = uwsgi_perl_magic,
 };
