@@ -3,66 +3,123 @@
 extern struct uwsgi_server uwsgi;
 struct uwsgi_python up;
 
+extern struct http_status_codes hsc[];
+
 #include <glob.h>
 
 extern PyTypeObject uwsgi_InputType;
 
-struct option uwsgi_python_options[] = {
-	{"wsgi-file", required_argument, 0, LONG_ARGS_WSGI_FILE},
-	{"file", required_argument, 0, LONG_ARGS_FILE_CONFIG},
-	{"eval", required_argument, 0, LONG_ARGS_EVAL_CONFIG},
-	{"module", required_argument, 0, 'w'},
-	{"callable", required_argument, 0, LONG_ARGS_CALLABLE},
-	{"test", required_argument, 0, 'j'},
-	{"home", required_argument, 0, 'H'},
-	{"virtualenv", required_argument, 0, 'H'},
-	{"venv", required_argument, 0, 'H'},
-	{"pyhome", required_argument, 0, 'H'},
-	{"pythonpath", required_argument, 0, LONG_ARGS_PYTHONPATH},
-	{"python-path", required_argument, 0, LONG_ARGS_PYTHONPATH},
-	{"pymodule-alias", required_argument, 0, LONG_ARGS_PYMODULE_ALIAS},
-	{"post-pymodule-alias", required_argument, 0, LONG_ARGS_POST_PYMODULE_ALIAS},
-	{"import", required_argument, 0, LONG_ARGS_PYIMPORT},
-	{"pyimport", required_argument, 0, LONG_ARGS_PYIMPORT},
-	{"py-import", required_argument, 0, LONG_ARGS_PYIMPORT},
-	{"python-import", required_argument, 0, LONG_ARGS_PYIMPORT},
-	{"pp", required_argument, 0, LONG_ARGS_PYTHONPATH},
-	{"pyargv", required_argument, 0, LONG_ARGS_PYARGV},
-	{"optimize", required_argument, 0, 'O'},
-	{"paste", required_argument, 0, LONG_ARGS_PASTE},
+void uwsgi_opt_pythonpath(char *opt, char *value, void *foobar) {
+
+	int i;
+	glob_t g;
+
+	if (glob(value, GLOB_MARK, NULL, &g)) {
+		uwsgi_string_new_list(&up.python_path, value);
+	}
+	else {
+		for (i = 0; i < (int) g.gl_pathc; i++) {
+	        	uwsgi_string_new_list(&up.python_path, g.gl_pathv[i]);
+	        }
+	}
+}
+
+void uwsgi_opt_pyshell(char *opt, char *value, void *foobar) {
+
+	uwsgi.honour_stdin = 1;
+	up.pyshell = 1;
+
+	if (!strcmp("pyshell-oneshot", opt)) {
+		up.pyshell_oneshot = 1;
+	}
+}
+
+void uwsgi_opt_pyrun(char *opt, char *value, void *foobar) {
+	uwsgi.honour_stdin = 1;
+	uwsgi.command_mode = 1;
+	up.pyrun = value;
+}
+
 #ifdef UWSGI_INI
-	{"ini-paste", required_argument, 0, LONG_ARGS_INI_PASTE},
+void uwsgi_opt_ini_paste(char *opt, char *value, void *foobar) {
+
+	uwsgi_opt_load_ini(opt, value, NULL);
+
+	if (value[0] != '/') {
+		up.paste = uwsgi_concat4("config:", uwsgi.cwd, "/", value);
+	}
+	else {
+		up.paste = uwsgi_concat2("config:", value);
+        }
+
+	if (!strcmp("ini-paste-logged", opt)) {
+		up.paste_logger = 1;
+	}
+	
+}
 #endif
-	{"catch-exceptions", no_argument, &up.catch_exceptions, 1},
-	{"ignore-script-name", no_argument, &up.ignore_script_name, 1},
-	{"pep3333-input", no_argument, &up.pep3333_input, 1},
-	{"reload-os-env", no_argument, &up.reload_os_env, 1},
-	{"no-site", no_argument, &Py_NoSiteFlag, 1},
 
-	{0, 0, 0, 0},
-};
+struct uwsgi_option uwsgi_python_options[] = {
+	{"wsgi-file", required_argument, 0, "load .wsgi file", uwsgi_opt_set_str, &up.file_config, 0},
+	{"file", required_argument, 0, "load .wsgi file", uwsgi_opt_set_str, &up.file_config, 0},
+	{"eval", required_argument, 0, "eval python code", uwsgi_opt_set_str, &up.eval, 0},
+	{"module", required_argument,'w', "load a WSGI module", uwsgi_opt_set_str, &up.wsgi_config, 0},
+	{"wsgi", required_argument, 'w', "load a WSGI module", uwsgi_opt_set_str, &up.wsgi_config, 0},
+	{"callable", required_argument, 0, "set default WSGI callable name", uwsgi_opt_set_str, &up.callable, 0},
+	{"test", required_argument, 'J', "test a mdule import", uwsgi_opt_set_str, &up.test_module, 0},
+	{"home", required_argument, 'H', "set PYTHONHOME/virtualenv", uwsgi_opt_set_str, &up.home, 0},
+	{"virtualenv", required_argument, 'H', "set PYTHONHOME/virtualenv", uwsgi_opt_set_str, &up.home, 0},
+	{"venv", required_argument, 'H', "set PYTHONHOME/virtualenv", uwsgi_opt_set_str, &up.home, 0},
+	{"pyhome", required_argument, 'H', "set PYTHONHOME/virtualenv", uwsgi_opt_set_str, &up.home, 0},
 
-struct uwsgi_help_item uwsgi_python_help[] = {
+	{"pythonpath", required_argument, 0, "add directory (or glob) to pythonpath", uwsgi_opt_pythonpath, NULL,  0},
+	{"python-path", required_argument, 0, "add directory (or glob) to pythonpath", uwsgi_opt_pythonpath, NULL, 0},
+	{"pp", required_argument, 0, "add directory (or glob) to pythonpath", uwsgi_opt_pythonpath, NULL, 0},
 
-{"module <module>" ,"name of python config module"},
-{"optimize <n>", "set python optimization level to <n>"},
-{"home <path>", "set python home/virtualenv"},
-{"pyhome <path>", "set python home/virtualenv"},
-{"virtualenv <path>", "set python home/virtualenv"},
-{"venv <path>", "set python home/virtualenv"},
-{"callable <callable>", "set the callable (default 'application')"},
-{"paste <config:/egg:>", "load applications using paste.deploy.loadapp()"},
-{"pythonpath <dir>", "add <dir> to PYTHONPATH"},
-{"python-path <dir>", "add <dir> to PYTHONPATH"},
-{"pp <dir>", "add <dir> to PYTHONPATH"},
-{"pyargv <args>", "assign args to python sys.argv"},
-{"wsgi-file <file>", "load the <file> wsgi file"},
-{"file <file>", "use python file instead of python module for configuration"},
-{"eval <code>", "evaluate code for app configuration"},
-{"ini-paste <inifile>", "path of ini config file that contains paste configuration"},
+	{"pymodule-alias", required_argument, 0, "add a python alias module", uwsgi_opt_add_string_list, &up.pymodule_alias, 0},
+	{"post-pymodule-alias", required_argument, 0, "add a python module alias after uwsgi module initialization", uwsgi_opt_add_string_list, &up.post_pymodule_alias, 0},
+
+	{"import", required_argument, 0, "import a python module", uwsgi_opt_add_string_list, &up.import_list, 0},
+	{"pyimport", required_argument, 0, "import a python module", uwsgi_opt_add_string_list, &up.import_list, 0},
+	{"py-import", required_argument, 0, "import a python module", uwsgi_opt_add_string_list, &up.import_list, 0},
+	{"python-import", required_argument, 0, "import a python module", uwsgi_opt_add_string_list, &up.import_list, 0},
+
+	{"shared-import", required_argument, 0, "import a python module in all of the processes", uwsgi_opt_add_string_list, &up.shared_import_list, 0},
+	{"shared-pyimport", required_argument, 0, "import a python module in all of the processes", uwsgi_opt_add_string_list, &up.shared_import_list, 0},
+	{"shared-py-import", required_argument, 0, "import a python module in all of the processes", uwsgi_opt_add_string_list, &up.shared_import_list, 0},
+	{"shared-python-import", required_argument, 0, "import a python module in all of the processes", uwsgi_opt_add_string_list, &up.shared_import_list, 0},
+
+	{"spooler-import", required_argument, 0, "import a python module in the spooler", uwsgi_opt_add_string_list, &up.spooler_import_list},
+	{"spooler-pyimport", required_argument, 0, "import a python module in the spooler", uwsgi_opt_add_string_list, &up.spooler_import_list},
+	{"spooler-py-import", required_argument, 0, "import a python module in the spooler", uwsgi_opt_add_string_list, &up.spooler_import_list},
+	{"spooler-python-import", required_argument, 0, "import a python module in the spooler", uwsgi_opt_add_string_list, &up.spooler_import_list},
+
+	{"pyargv", required_argument, 0, "manually set sys.argv", uwsgi_opt_set_str, &up.argv, 0},
+	{"optimize", required_argument, 'O', "set python optimization level", uwsgi_opt_set_int, &up.optimize, 0},
+
+	{"paste", required_argument, 0, "load a paste.deploy config file", uwsgi_opt_set_str, &up.paste, 0},
+	{"paste-logger", no_argument, 0, "enable paste fileConfig logger", uwsgi_opt_true, &up.paste_logger, 0},
 
 
- { 0, 0},
+	{"web3", required_argument, 0, "load a web3 app", uwsgi_opt_set_str, &up.web3, 0},
+	{"pump", required_argument, 0, "load a pump app", uwsgi_opt_set_str, &up.pump, 0},
+	{"wsgi-lite", required_argument, 0, "load a wsgi-lite app", uwsgi_opt_set_str, &up.wsgi_lite, 0},
+#ifdef UWSGI_INI
+	{"ini-paste", required_argument, 0, "load a paste.deploy config file containing uwsgi section", uwsgi_opt_ini_paste, NULL, UWSGI_OPT_IMMEDIATE},
+	{"ini-paste-logged", required_argument, 0, "load a paste.deploy config file containing uwsgi section (load loggers too)", uwsgi_opt_ini_paste, NULL, UWSGI_OPT_IMMEDIATE},
+#endif
+	{"reload-os-env", no_argument, 0, "force reload of os.environ at each request", uwsgi_opt_true, &up.reload_os_env, 0},
+#ifndef UWSGI_PYPY
+	{"no-site", no_argument, 0, "do not import site module", uwsgi_opt_true, &Py_NoSiteFlag, 0},
+#endif
+	{"pyshell", no_argument, 0, "run an interactive python shell in the uWSGI environment", uwsgi_opt_pyshell, NULL, 0},
+	{"pyshell-oneshot", no_argument, 0, "run an interactive python shell in the uWSGI environment (one-shot variant)", uwsgi_opt_pyshell, NULL, 0},
+
+	{"python", required_argument, 0, "run a python script in the uWSGI environment", uwsgi_opt_pyrun, NULL, 0},
+	{"py", required_argument, 0, "run a python script in the uWSGI environment", uwsgi_opt_pyrun, NULL, 0},
+	{"pyrun", required_argument, 0, "run a python script in the uWSGI environment", uwsgi_opt_pyrun, NULL, 0},
+
+	{0, 0, 0, 0, 0, 0, 0},
 };
 
 /* this routine will be called after each fork to reinitialize the various locks */
@@ -78,19 +135,19 @@ void uwsgi_python_pthread_child(void) {
 	pthread_mutex_init(&up.lock_pyloaders, NULL);
 }
 
-// fake method
-PyMethodDef null_methods[] = {
-	{NULL, NULL},
-};
-
 PyMethodDef uwsgi_spit_method[] = { {"uwsgi_spit", py_uwsgi_spit, METH_VARARGS, ""} };
 PyMethodDef uwsgi_write_method[] = { {"uwsgi_write", py_uwsgi_write, METH_VARARGS, ""} };
 
 int uwsgi_python_init() {
 
+#ifndef UWSGI_PYPY
 	char *pyversion = strchr(Py_GetVersion(), '\n');
-        uwsgi_log("Python version: %.*s %s\n", pyversion-Py_GetVersion(), Py_GetVersion(), Py_GetCompiler()+1);
+        uwsgi_log_initial("Python version: %.*s %s\n", pyversion-Py_GetVersion(), Py_GetVersion(), Py_GetCompiler()+1);
+#else
+	uwsgi_log_initial("PyPy version: %s\n", PYPY_VERSION);
+#endif
 
+#ifndef UWSGI_PYPY
 	if (up.home != NULL) {
 #ifdef PYTHREE
 		wchar_t *wpyhome;
@@ -119,9 +176,12 @@ int uwsgi_python_init() {
 #endif
 
 
+	Py_OptimizeFlag = up.optimize;
+
 	Py_Initialize();
 
-	Py_OptimizeFlag = up.optimize;
+
+#endif
 
 	up.wsgi_spitout = PyCFunction_New(uwsgi_spit_method, NULL);
 	up.wsgi_writeout = PyCFunction_New(uwsgi_write_method, NULL);
@@ -136,7 +196,7 @@ int uwsgi_python_init() {
         up.reset_ts = simple_reset_ts;
 	
 
-	uwsgi_log("Python main interpreter initialized at %p\n", up.main_thread);
+	uwsgi_log_initial("Python main interpreter initialized at %p\n", up.main_thread);
 
 	return 1;
 
@@ -144,6 +204,7 @@ int uwsgi_python_init() {
 
 void uwsgi_python_reset_random_seed() {
 
+#ifndef UWSGI_PYPY
 	PyObject *random_module, *random_dict, *random_seed;
 
         // reinitialize the random seed (thanks Jonas Borgström)
@@ -163,23 +224,81 @@ void uwsgi_python_reset_random_seed() {
                         }
                 }
         }
+#endif
+}
+
+
+void uwsgi_python_atexit() {
+
+	// if hijacked do not run atexit hooks
+	if (uwsgi.workers[uwsgi.mywid].hijacked)
+		return;
+
+	// if busy do not run atexit hooks
+	if (uwsgi.workers[uwsgi.mywid].busy)
+		return;
+
+#ifdef UWSGI_ASYNC
+	// managing atexit in async mode is a real pain...skip it for now
+	if (uwsgi.async > 1)
+		return;
+#endif
+
+	// this time we use this higher level function
+	// as this code can be executed in a signal handler
+
+	if (!Py_IsInitialized()) {
+		return;
+	}
+
+	if (uwsgi.has_threads)
+		PyGILState_Ensure();
+	// no need to worry about freeing memory
+	PyObject *uwsgi_dict = get_uwsgi_pydict("uwsgi");
+	if (uwsgi_dict) {
+		PyObject *ae = PyDict_GetItemString(uwsgi_dict, "atexit");
+		if (ae) {
+			python_call(ae, PyTuple_New(0), 0, NULL);
+		}
+	}
+
+	// this part is a 1:1 copy of mod_wsgi 3.x
+        // it is required to fix some atexit bug with python 3
+	// and to shutdown useless threads complaints
+	PyObject *module = PyImport_ImportModule("atexit");
+	Py_XDECREF(module);
+
+	if (uwsgi.threads > 1) {
+		if (!PyImport_AddModule("dummy_threading"))
+			PyErr_Clear();
+	}
+
+	Py_Finalize();
 }
 
 void uwsgi_python_post_fork() {
 
+#ifdef UWSGI_SPOOLER
+	if (uwsgi.i_am_a_spooler) {
+		UWSGI_GET_GIL
+	}	
+#endif
+
 	uwsgi_python_reset_random_seed();
 
+#ifndef UWSGI_PYPY
 #ifdef UWSGI_EMBEDDED
 	// call the post_fork_hook
 	PyObject *uwsgi_dict = get_uwsgi_pydict("uwsgi");
 	if (uwsgi_dict) {
 		PyObject *pfh = PyDict_GetItemString(uwsgi_dict, "post_fork_hook");
 		if (pfh) {
-			python_call(pfh, PyTuple_New(0), 0);
+			python_call(pfh, PyTuple_New(0), 0, NULL);
 		}
 	}
 	PyErr_Clear();
 #endif
+#endif
 
 UWSGI_RELEASE_GIL
 
@@ -187,6 +306,7 @@ UWSGI_RELEASE_GIL
 
 PyObject *uwsgi_pyimport_by_filename(char *name, char *filename) {
 
+#ifndef UWSGI_PYPY
 	FILE *pyfile;
 	struct _node *py_file_node = NULL;
 	PyObject *py_compiled_node, *py_file_module;
@@ -195,7 +315,7 @@ PyObject *uwsgi_pyimport_by_filename(char *name, char *filename) {
 	char *real_filename = filename;
 
 
-	if (strncmp(filename, "http://", 7) && strncmp(filename, "sym://", 6)) {
+	if (!uwsgi_check_scheme(filename)) {
 
 		pyfile = fopen(filename, "r");
 		if (!pyfile) {
@@ -216,7 +336,6 @@ PyObject *uwsgi_pyimport_by_filename(char *name, char *filename) {
 			if (!pyfile) {
 				uwsgi_error_open(real_filename);
 				free(real_filename);
-				fclose(pyfile);
 				return NULL;
 			}
 		}
@@ -225,7 +344,8 @@ PyObject *uwsgi_pyimport_by_filename(char *name, char *filename) {
 		if (!py_file_node) {
 			PyErr_Print();
 			uwsgi_log("failed to parse file %s\n", real_filename);
-			free(real_filename);
+			if (is_a_package)
+				free(real_filename);
 			fclose(pyfile);
 			return NULL;
 		}
@@ -271,23 +391,19 @@ PyObject *uwsgi_pyimport_by_filename(char *name, char *filename) {
 	}
 
 	return py_file_module;
+#else
+	return NULL;
+#endif
 
 }
 
-
 void init_uwsgi_vars() {
 
-	int i;
 	PyObject *pysys, *pysys_dict, *pypath;
 
 	PyObject *modules = PyImport_GetModuleDict();
 	PyObject *tmp_module;
 
-#ifdef UWSGI_MINTERPRETERS
-	char venv_version[15];
-	PyObject *site_module;
-#endif
-
 	/* add cwd to pythonpath */
 	pysys = PyImport_ImportModule("sys");
 	if (!pysys) {
@@ -310,39 +426,6 @@ void init_uwsgi_vars() {
 		exit(1);
 	}
 
-#ifdef UWSGI_MINTERPRETERS
-	// simulate a pythonhome directive
-	if (uwsgi.wsgi_req->pyhome_len > 0) {
-
-		PyObject *venv_path = UWSGI_PYFROMSTRINGSIZE(uwsgi.wsgi_req->pyhome, uwsgi.wsgi_req->pyhome_len);
-
-#ifdef UWSGI_DEBUG
-		uwsgi_debug("setting dynamic virtualenv to %.*s\n", uwsgi.wsgi_req->pyhome_len, uwsgi.wsgi_req->pyhome);
-#endif
-
-		PyDict_SetItemString(pysys_dict, "prefix", venv_path);
-		PyDict_SetItemString(pysys_dict, "exec_prefix", venv_path);
-
-		venv_version[14] = 0;
-		if (snprintf(venv_version, 15, "/lib/python%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION) == -1) {
-			return;
-		}
-
-		// check here
-		PyString_Concat(&venv_path, PyString_FromString(venv_version));
-
-		if (PyList_Insert(pypath, 0, venv_path)) {
-			PyErr_Print();
-		}
-
-		site_module = PyImport_ImportModule("site");
-		if (site_module) {
-			PyImport_ReloadModule(site_module);
-		}
-
-	}
-#endif
-
 	if (PyList_Insert(pypath, 0, UWSGI_PYFROMSTRING(".")) != 0) {
 		PyErr_Print();
 	}
@@ -359,9 +442,10 @@ void init_uwsgi_vars() {
 		uppp = uppp->next;
 	}
 
-	for (i = 0; i < up.pymodule_alias_cnt; i++) {
+	struct uwsgi_string_list *uppma = up.pymodule_alias;
+	while(uppma) {
 		// split key=value
-		char *value = strchr(up.pymodule_alias[i], '=');
+		char *value = strchr(uppma->value, '=');
 		if (!value) {
 			uwsgi_log("invalid pymodule-alias syntax\n");
 			continue;
@@ -375,39 +459,43 @@ void init_uwsgi_vars() {
 				exit(1);
 			}
 
-			PyDict_SetItemString(modules, up.pymodule_alias[i], tmp_module);
+			PyDict_SetItemString(modules, uppma->value, tmp_module);
 		}
 		else {
 			// this is a filepath that need to be mapped
-			tmp_module = uwsgi_pyimport_by_filename(up.pymodule_alias[i], value + 1);
+			tmp_module = uwsgi_pyimport_by_filename(uppma->value, value + 1);
 			if (!tmp_module) {
 				PyErr_Print();
 				exit(1);
 			}
 		}
-		uwsgi_log("mapped virtual pymodule \"%s\" to real pymodule \"%s\"\n", up.pymodule_alias[i], value + 1);
+		uwsgi_log("mapped virtual pymodule \"%s\" to real pymodule \"%s\"\n", uppma->value, value + 1);
 		// reset original value
 		value[0] = '=';
+
+		uppma = uppma->next;
 	}
 
 }
 
 
 
+PyDoc_STRVAR(uwsgi_py_doc, "uWSGI api module.");
+
+
 #ifdef PYTHREE
 static PyModuleDef uwsgi_module3 = {
 	PyModuleDef_HEAD_INIT,
 	"uwsgi",
-	NULL,
+	uwsgi_py_doc,
 	-1,
-	null_methods,
+	NULL,
 };
 PyObject *init_uwsgi3(void) {
 	return PyModule_Create(&uwsgi_module3);
 }
 #endif
 
-
 #ifdef UWSGI_EMBEDDED
 void init_uwsgi_embedded_module() {
 	PyObject *new_uwsgi_module, *zero;
@@ -424,24 +512,30 @@ void init_uwsgi_embedded_module() {
 	}
 
 
-
 #ifdef PYTHREE
 	PyImport_AppendInittab("uwsgi", init_uwsgi3);
 	new_uwsgi_module = PyImport_AddModule("uwsgi");
 #else
-	new_uwsgi_module = Py_InitModule("uwsgi", null_methods);
+	new_uwsgi_module = Py_InitModule3("uwsgi", NULL, uwsgi_py_doc);
 #endif
 	if (new_uwsgi_module == NULL) {
 		uwsgi_log("could not initialize the uwsgi python module\n");
 		exit(1);
 	}
 
+
+
+	Py_INCREF((PyObject *) &uwsgi_InputType);
+
 	up.embedded_dict = PyModule_GetDict(new_uwsgi_module);
 	if (!up.embedded_dict) {
 		uwsgi_log("could not get uwsgi module __dict__\n");
 		exit(1);
 	}
 
+	// just for safety
+	Py_INCREF(up.embedded_dict);
+
 	if (PyDict_SetItemString(up.embedded_dict, "version", PyString_FromString(UWSGI_VERSION))) {
 		PyErr_Print();
 		exit(1);
@@ -461,6 +555,7 @@ void init_uwsgi_embedded_module() {
 	}
 
 
+
 	if (PyDict_SetItemString(up.embedded_dict, "hostname", PyString_FromStringAndSize(uwsgi.hostname, uwsgi.hostname_len))) {
 		PyErr_Print();
 		exit(1);
@@ -480,6 +575,31 @@ void init_uwsgi_embedded_module() {
 		}
 	}
 
+#ifdef UWSGI_SPOOLER
+	if (uwsgi.spoolers) {
+		int sc = 0;
+		struct uwsgi_spooler *uspool = uwsgi.spoolers;
+		while(uspool) { sc++; uspool = uspool->next;}
+
+		PyObject *py_spooler_tuple = PyTuple_New(sc);
+
+		uspool = uwsgi.spoolers;
+		sc = 0;
+
+		while(uspool) {
+			PyTuple_SetItem(py_spooler_tuple, sc, PyString_FromString(uspool->dir));
+			sc++;
+			uspool = uspool->next;
+		}
+
+		if (PyDict_SetItemString(up.embedded_dict, "spoolers", py_spooler_tuple)) {
+                	PyErr_Print();
+                	exit(1);
+        	}
+	}
+#endif
+
+
 
 	if (PyDict_SetItemString(up.embedded_dict, "SPOOL_RETRY", PyInt_FromLong(-1))) {
 		PyErr_Print();
@@ -517,13 +637,21 @@ void init_uwsgi_embedded_module() {
 			exit(1);
 		}
 	}
+	else {
+		PyDict_SetItemString(up.embedded_dict, "loop", Py_None);
+	}
 
 	PyObject *py_opt_dict = PyDict_New();
 	for (i = 0; i < uwsgi.exported_opts_cnt; i++) {
 		if (PyDict_Contains(py_opt_dict, PyString_FromString(uwsgi.exported_opts[i]->key))) {
 			PyObject *py_opt_item = PyDict_GetItemString(py_opt_dict, uwsgi.exported_opts[i]->key);
 			if (PyList_Check(py_opt_item)) {
-				PyList_Append(py_opt_item, PyString_FromString(uwsgi.exported_opts[i]->value));
+				if (uwsgi.exported_opts[i]->value == NULL) {
+					PyList_Append(py_opt_item, Py_True);
+				}
+				else {
+					PyList_Append(py_opt_item, PyString_FromString(uwsgi.exported_opts[i]->value));
+				}
 			}
 			else {
 				PyObject *py_opt_list = PyList_New(0);
@@ -553,6 +681,23 @@ void init_uwsgi_embedded_module() {
 		exit(1);
 	}
 
+	PyObject *py_magic_table = PyDict_New();
+	uint8_t mtk;
+	for (i = 0; i <= 0xff; i++) {
+		// a bit of magic :P
+		mtk = i;
+                if (uwsgi.magic_table[i]) {
+			if (uwsgi.magic_table[i][0] != 0) {
+				PyDict_SetItem(py_magic_table, PyString_FromStringAndSize((char *) &mtk, 1), PyString_FromString(uwsgi.magic_table[i]));
+			}
+		}
+        }
+
+	if (PyDict_SetItemString(up.embedded_dict, "magic_table", py_magic_table)) {
+		PyErr_Print();
+		exit(1);
+	}
+
 #ifdef UNBIT
 	if (PyDict_SetItemString(up.embedded_dict, "unbit", Py_True)) {
 #else
@@ -577,11 +722,6 @@ void init_uwsgi_embedded_module() {
 		exit(1);
 	}
 
-	if (PyDict_SetItemString(up.embedded_dict, "fastfuncs", PyList_New(256))) {
-		PyErr_Print();
-		exit(1);
-	}
-
 
 	if (PyDict_SetItemString(up.embedded_dict, "applications", Py_None)) {
 		PyErr_Print();
@@ -612,16 +752,10 @@ void init_uwsgi_embedded_module() {
 		exit(1);
 	}
 
-	up.fastfuncslist = PyDict_GetItemString(up.embedded_dict, "fastfuncs");
-	if (!up.fastfuncslist) {
-		PyErr_Print();
-		exit(1);
-	}
-
 	init_uwsgi_module_advanced(new_uwsgi_module);
 
 #ifdef UWSGI_SPOOLER
-	if (uwsgi.spool_dir != NULL) {
+	if (uwsgi.spoolers) {
 		init_uwsgi_module_spooler(new_uwsgi_module);
 	}
 #endif
@@ -631,9 +765,19 @@ void init_uwsgi_embedded_module() {
 		init_uwsgi_module_sharedarea(new_uwsgi_module);
 	}
 
-	init_uwsgi_module_cache(new_uwsgi_module);
+	if (uwsgi.cache_max_items > 0) {
+		init_uwsgi_module_cache(new_uwsgi_module);
+	}
 
-	init_uwsgi_module_queue(new_uwsgi_module);
+	if (uwsgi.queue_size > 0) {
+		init_uwsgi_module_queue(new_uwsgi_module);
+	}
+
+#ifdef UWSGI_SNMP
+	if (uwsgi.snmp) {
+		init_uwsgi_module_snmp(new_uwsgi_module);
+	}
+#endif
 
 	if (up.extension) {
 		up.extension();
@@ -672,94 +816,37 @@ int uwsgi_python_magic(char *mountpoint, char *lazy) {
 
 }
 
-int uwsgi_python_manage_options(int i, char *optarg) {
+int uwsgi_python_mount_app(char *mountpoint, char *app, int regexp) {
 
-	glob_t g;
-	int j;
+	int id;
 
-	switch (i) {
-	case 'w':
-		up.wsgi_config = optarg;
-		return 1;
-	case LONG_ARGS_WSGI_FILE:
-	case LONG_ARGS_FILE_CONFIG:
-		up.file_config = optarg;
-		return 1;
-	case LONG_ARGS_EVAL_CONFIG:
-		up.eval = optarg;
-		return 1;
-	case LONG_ARGS_PYMODULE_ALIAS:
-		if (up.pymodule_alias_cnt < MAX_PYMODULE_ALIAS) {
-			up.pymodule_alias[up.pymodule_alias_cnt] = optarg;
-			up.pymodule_alias_cnt++;
+	if (strchr(app, ':') || uwsgi_endswith(app, ".py") || uwsgi_endswith(app, ".wsgi")) {
+		uwsgi.wsgi_req->appid = mountpoint;
+		uwsgi.wsgi_req->appid_len = strlen(mountpoint);
+		if (uwsgi.single_interpreter) {
+			id = init_uwsgi_app(LOADER_MOUNT, app, uwsgi.wsgi_req, up.main_thread, PYTHON_APP_TYPE_WSGI);
 		}
 		else {
-			uwsgi_log("you can specify at most %d --pymodule-alias options\n", MAX_PYMODULE_ALIAS);
+			id = init_uwsgi_app(LOADER_MOUNT, app, uwsgi.wsgi_req, NULL, PYTHON_APP_TYPE_WSGI);
 		}
-		return 1;
-	case LONG_ARGS_PYIMPORT:
-		uwsgi_string_new_list(&up.import_list, optarg);
-		return 1;
-	case LONG_ARGS_POST_PYMODULE_ALIAS:
-		uwsgi_string_new_list(&up.post_pymodule_alias, optarg);
-		return 1;
-	case LONG_ARGS_PYTHONPATH:
-		if (glob(optarg, GLOB_MARK, NULL, &g)) {
-			uwsgi_string_new_list(&up.python_path, optarg);
-		}
-		else {
-			for (j = 0; j < (int) g.gl_pathc; j++) {
-				uwsgi_string_new_list(&up.python_path, g.gl_pathv[j]);
+
+#ifdef UWSGI_PCRE
+	int i;
+	if (regexp && id != -1) {
+		struct uwsgi_app *ua = &uwsgi_apps[id];
+		uwsgi_regexp_build(mountpoint, &ua->pattern, &ua->pattern_extra);
+		if (uwsgi.mywid == 0) {
+			for(i=1;i<=uwsgi.numproc;i++) {
+				uwsgi.workers[i].apps[id].pattern = ua->pattern;
+				uwsgi.workers[i].apps[id].pattern_extra = ua->pattern_extra;
 			}
 		}
-		return 1;
-	case LONG_ARGS_PYARGV:
-		up.argv = optarg;
-		return 1;
-	case 'j':
-		up.test_module = optarg;
-		return 1;
-	case 'H':
-		up.home = optarg;
-		return 1;
-	case 'O':
-		up.optimize = atoi(optarg);
-		return 1;
-	case LONG_ARGS_CALLABLE:
-		up.callable = optarg;
-		return 1;
-
-
-#ifdef UWSGI_INI
-	case LONG_ARGS_INI_PASTE:
-		uwsgi.ini = optarg;
-		if (uwsgi.ini[0] != '/') {
-			up.paste = uwsgi_concat4("config:", uwsgi.cwd, "/", uwsgi.ini);
-		}
-		else {
-			up.paste = uwsgi_concat2("config:", uwsgi.ini);
-		}
-		return 1;
+	}
 #endif
-	case LONG_ARGS_PASTE:
-		up.paste = optarg;
-		return 1;
-
-
 
+		return id;
 	}
-
-	return 0;
-}
-
-int uwsgi_python_mount_app(char *mountpoint, char *app) {
-
-	uwsgi.wsgi_req->script_name = mountpoint;
-	uwsgi.wsgi_req->script_name_len = strlen(mountpoint);
-	if (uwsgi.single_interpreter) {
-		return init_uwsgi_app(LOADER_MOUNT, app, uwsgi.wsgi_req, up.main_thread);
-	}
-	return init_uwsgi_app(LOADER_MOUNT, app, uwsgi.wsgi_req, NULL);
+	return -1;
 
 }
 
@@ -775,11 +862,17 @@ char *uwsgi_pythonize(char *orig) {
 	else if (!strncmp(name, "http://", 7)) {
 		name+=7;
 	}
+	else if (!strncmp(name, "data://", 7)) {
+		name+=7;
+	}
 
 	len = strlen(name);
 	for(i=0;i<len;i++) {
-		if (name[i] == '/') {
-			name[i] = '.';
+		if (name[i] == '.') {
+			name[i] = '_';
+		}
+		else if (name[i] == '/') {
+			name[i] = '_';
 		}
 	}
 
@@ -792,24 +885,46 @@ char *uwsgi_pythonize(char *orig) {
 
 }
 
-void uwsgi_python_init_apps() {
+void uwsgi_python_spooler_init(void) {
 
+	struct uwsgi_string_list *upli = up.spooler_import_list;
 
-	if (uwsgi.async > 1) {
-		up.current_recursion_depth = uwsgi_malloc(sizeof(int)*uwsgi.async);
-        	up.current_frame = uwsgi_malloc(sizeof(struct _frame)*uwsgi.async);
-	}
+	UWSGI_GET_GIL
+
+        while(upli) {
+                if (strchr(upli->value, '/') || uwsgi_endswith(upli->value, ".py")) {
+                        uwsgi_pyimport_by_filename(uwsgi_pythonize(upli->value), upli->value);
+                }
+                else {
+                        if (PyImport_ImportModule(upli->value) == NULL) {
+                                PyErr_Print();
+                        }
+                }
+                upli = upli->next;
+        }
+
+	UWSGI_RELEASE_GIL
+	
+
+}
+
+// this hook will be executed by master (or worker1 when master is not requested, so COW is in place)
+void uwsgi_python_preinit_apps() {
 
 	init_pyargv();
 
-#ifdef UWSGI_MINTERPRETERS
+#ifndef UWSGI_PYPY
+#ifdef UWSGI_EMBEDDED
         init_uwsgi_embedded_module();
 #endif
+#endif
 
 #ifdef __linux__
-#ifndef PYTHREE
+#ifndef UWSGI_PYPY
+#ifdef UWSGI_EMBEDDED
 	uwsgi_init_symbol_import();
 #endif
+#endif
 #endif
 
         if (up.test_module != NULL) {
@@ -821,6 +936,34 @@ void uwsgi_python_init_apps() {
 
         init_uwsgi_vars();
 
+	// load shared imports
+	struct uwsgi_string_list *upli = up.shared_import_list;
+	while(upli) {
+		if (strchr(upli->value, '/') || uwsgi_endswith(upli->value, ".py")) {
+			uwsgi_pyimport_by_filename(uwsgi_pythonize(upli->value), upli->value);
+		}
+		else {
+			if (PyImport_ImportModule(upli->value) == NULL) {
+				PyErr_Print();
+			}
+		}
+		upli = upli->next;
+	}
+
+}
+
+void uwsgi_python_init_apps() {
+
+	struct http_status_codes *http_sc;
+
+	// prepare for stack suspend/resume
+	if (uwsgi.async > 1) {
+		up.current_recursion_depth = uwsgi_malloc(sizeof(int)*uwsgi.async);
+#ifndef UWSGI_PYPY
+        	up.current_frame = uwsgi_malloc(sizeof(struct _frame)*uwsgi.async);
+#endif
+	}
+
         // setup app loaders
 #ifdef UWSGI_MINTERPRETERS
         up.loaders[LOADER_DYN] = uwsgi_dyn_loader;
@@ -885,23 +1028,48 @@ void uwsgi_python_init_apps() {
 
 
 	if (up.wsgi_config != NULL) {
-		init_uwsgi_app(LOADER_UWSGI, up.wsgi_config, uwsgi.wsgi_req, up.main_thread);
+		init_uwsgi_app(LOADER_UWSGI, up.wsgi_config, uwsgi.wsgi_req, up.main_thread, PYTHON_APP_TYPE_WSGI);
 	}
 
 	if (up.file_config != NULL) {
-		init_uwsgi_app(LOADER_FILE, up.file_config, uwsgi.wsgi_req, up.main_thread);
+		init_uwsgi_app(LOADER_FILE, up.file_config, uwsgi.wsgi_req, up.main_thread, PYTHON_APP_TYPE_WSGI);
 	}
 	if (up.paste != NULL) {
-		init_uwsgi_app(LOADER_PASTE, up.paste, uwsgi.wsgi_req, up.main_thread);
+		init_uwsgi_app(LOADER_PASTE, up.paste, uwsgi.wsgi_req, up.main_thread, PYTHON_APP_TYPE_WSGI);
 	}
 	if (up.eval != NULL) {
-		init_uwsgi_app(LOADER_EVAL, up.eval, uwsgi.wsgi_req, up.main_thread);
+		init_uwsgi_app(LOADER_EVAL, up.eval, uwsgi.wsgi_req, up.main_thread, PYTHON_APP_TYPE_WSGI);
+	}
+	if (up.web3 != NULL) {
+		init_uwsgi_app(LOADER_UWSGI, up.web3, uwsgi.wsgi_req, up.main_thread, PYTHON_APP_TYPE_WEB3);
+	}
+	if (up.pump != NULL) {
+		init_uwsgi_app(LOADER_UWSGI, up.pump, uwsgi.wsgi_req, up.main_thread, PYTHON_APP_TYPE_PUMP);
+		// filling http status codes
+        	for (http_sc = hsc; http_sc->message != NULL; http_sc++) {
+                	http_sc->message_size = (int) strlen(http_sc->message);
+        	}
+	}
+	if (up.wsgi_lite != NULL) {
+		init_uwsgi_app(LOADER_UWSGI, up.wsgi_lite, uwsgi.wsgi_req, up.main_thread, PYTHON_APP_TYPE_WSGI_LITE);
 	}
 
 	if (uwsgi.profiler) {
+#ifndef UWSGI_PYPY
 		if (!strcmp(uwsgi.profiler, "pycall")) {
 			PyEval_SetProfile(uwsgi_python_profiler_call, NULL);
 		}
+#endif
+	}
+
+	PyObject *uwsgi_dict = get_uwsgi_pydict("uwsgi");
+        if (uwsgi_dict) {
+                up.after_req_hook = PyDict_GetItemString(uwsgi_dict, "after_req_hook");
+                if (up.after_req_hook) {
+			Py_INCREF(up.after_req_hook);
+			up.after_req_hook_args = PyTuple_New(0);
+			Py_INCREF(up.after_req_hook_args);
+		}
 	}
 
 }
@@ -948,11 +1116,16 @@ void uwsgi_python_enable_threads() {
 	up.gil_get = gil_real_get;
 	up.gil_release = gil_real_release;
 
+	up.swap_ts = simple_threaded_swap_ts;
+	up.reset_ts = simple_threaded_reset_ts;
+
 	if (uwsgi.threads > 1) {
 		up.swap_ts = threaded_swap_ts;
 		up.reset_ts = threaded_reset_ts;
 	}
+
 	uwsgi_log("threads support enabled\n");
+	
 
 }
 
@@ -963,59 +1136,39 @@ void uwsgi_python_init_thread(int core_id) {
 	pts = PyThreadState_New(up.main_thread->interp);
 	pthread_setspecific(up.upt_save_key, (void *) pts);
 	pthread_setspecific(up.upt_gil_key, (void *) pts);
+#ifdef UWSGI_DEBUG
+	uwsgi_log("python ThreadState %d = %p\n", core_id, pts);
+#endif
+	UWSGI_GET_GIL;
+	// call threading.currentThread (taken from mod_wsgi, but removes DECREFs as thread in uWSGI are fixed)
+	PyObject *threading_module = PyImport_ImportModule("threading");
+        if (threading_module) {
+        	PyObject *threading_module_dict = PyModule_GetDict(threading_module);
+                if (threading_module_dict) {
+#ifdef PYTHREE
+			PyObject *threading_current = PyDict_GetItemString(threading_module_dict, "current_thread");
+#else
+			PyObject *threading_current = PyDict_GetItemString(threading_module_dict, "currentThread");
+#endif
+                        if (threading_current) {
+                                PyObject *current_thread = PyEval_CallObject(threading_current, (PyObject *)NULL);
+                                if (!current_thread) {
+					// ignore the error
+                                        PyErr_Clear();
+                                }
+				else {
+					PyObject_SetAttrString(current_thread, "name", PyString_FromFormat("uWSGIWorker%dCore%d", uwsgi.mywid, core_id));
+					Py_INCREF(current_thread);
+				}
+                        }
+                }
+        }
+	UWSGI_RELEASE_GIL;
+	
 
 }
 
-int uwsgi_python_xml(char *node, char *content) {
-
-	PyThreadState *interpreter = NULL;
-
-	if (uwsgi.single_interpreter) {
-		interpreter = up.main_thread;
-	}
-
-	if (!strcmp("script", node)) {
-		return init_uwsgi_app(LOADER_UWSGI, content, uwsgi.wsgi_req, interpreter);
-	}
-	else if (!strcmp("file", node)) {
-		return init_uwsgi_app(LOADER_FILE, content, uwsgi.wsgi_req, interpreter);
-	}
-	else if (!strcmp("eval", node)) {
-		return init_uwsgi_app(LOADER_EVAL, content, uwsgi.wsgi_req, interpreter);
-	}
-	else if (!strcmp("wsgi", node)) {
-		return init_uwsgi_app(LOADER_EVAL, content, uwsgi.wsgi_req, interpreter);
-	}
-	else if (!strcmp("module", node)) {
-		uwsgi.wsgi_req->module = content;
-		uwsgi.wsgi_req->module_len = strlen(content);
-		uwsgi.wsgi_req->callable = strchr(uwsgi.wsgi_req->module, ':');
-		if (uwsgi.wsgi_req->callable) {
-			uwsgi.wsgi_req->callable[0] = 0;
-			uwsgi.wsgi_req->callable++;
-			uwsgi.wsgi_req->callable_len = strlen(uwsgi.wsgi_req->callable);
-			uwsgi.wsgi_req->module_len = strlen(uwsgi.wsgi_req->module);
-			return init_uwsgi_app(LOADER_DYN, uwsgi.wsgi_req, uwsgi.wsgi_req, interpreter);
-		}
-		else {
-			return init_uwsgi_app(LOADER_UWSGI, content, uwsgi.wsgi_req, interpreter);
-		}
-		return 1;
-	}
-	else if (!strcmp("pyhome", node)) {
-		uwsgi.wsgi_req->pyhome = content;
-		uwsgi.wsgi_req->pyhome_len = strlen(content);
-		return 1;
-	}
-	else if (!strcmp("callable", node)) {
-		uwsgi.wsgi_req->callable = content;
-		uwsgi.wsgi_req->callable_len = strlen(content);
-		return init_uwsgi_app(LOADER_DYN, uwsgi.wsgi_req, uwsgi.wsgi_req, interpreter);
-	}
-
-	return 0;
-}
-
+#ifndef UWSGI_PYPY
 void uwsgi_python_suspend(struct wsgi_request *wsgi_req) {
 
 	PyThreadState *tstate = PyThreadState_GET();
@@ -1029,6 +1182,57 @@ void uwsgi_python_suspend(struct wsgi_request *wsgi_req) {
 		up.current_main_frame = tstate->frame;
 	}
 
+}
+#endif
+
+char *uwsgi_python_code_string(char *id, char *code, char *function, char *key, uint16_t keylen) {
+
+	PyObject *cs_module = NULL;
+	PyObject *cs_dict = NULL;
+
+	UWSGI_GET_GIL;
+
+	cs_module = PyImport_ImportModule(id);
+	if (!cs_module) {
+		PyErr_Clear();
+		cs_module = uwsgi_pyimport_by_filename(id, code);
+	}
+
+	if (!cs_module) {
+		UWSGI_RELEASE_GIL;
+		return NULL;
+	}
+
+	cs_dict = PyModule_GetDict(cs_module);
+	if (!cs_dict) {
+		PyErr_Print();
+		UWSGI_RELEASE_GIL;
+		return NULL;
+	}
+	
+	PyObject *func = PyDict_GetItemString(cs_dict, function);
+	if (!func) {
+		uwsgi_log("function %s not available in %s\n", function, code);
+		PyErr_Print();
+		UWSGI_RELEASE_GIL;
+		return NULL;
+	}
+
+	PyObject *args = PyTuple_New(1);
+
+	PyTuple_SetItem(args, 0, PyString_FromStringAndSize(key, keylen));
+
+	PyObject *ret = python_call(func, args, 0, NULL);
+	Py_DECREF(args);
+	if (ret && PyString_Check(ret)) {
+		char *val = PyString_AsString(ret);
+		UWSGI_RELEASE_GIL;
+		return val;
+	}
+
+	UWSGI_RELEASE_GIL;
+	return NULL;
+	
 }
 
 int uwsgi_python_signal_handler(uint8_t sig, void *handler) {
@@ -1046,9 +1250,10 @@ int uwsgi_python_signal_handler(uint8_t sig, void *handler) {
 
 	PyTuple_SetItem(args, 0, PyInt_FromLong(sig));
 
-	ret = python_call(handler, args, 0);
-
+	ret = python_call(handler, args, 0, NULL);
+	Py_DECREF(args);
 	if (ret) {
+		Py_DECREF(ret);
 		UWSGI_RELEASE_GIL;
 		return 0;
 	}
@@ -1058,38 +1263,44 @@ clear:
 	return -1;
 }
 
-uint16_t uwsgi_python_rpc(void *func, uint8_t argc, char **argv, char *buffer) {
+uint16_t uwsgi_python_rpc(void *func, uint8_t argc, char **argv, uint16_t argvs[], char *buffer) {
+
+	UWSGI_GET_GIL;
 
 	uint8_t i;
-	PyObject *pyargs = PyTuple_New(argc);
-	PyObject *ret;
 	char *rv;
 	size_t rl;
 
+	PyObject *pyargs = PyTuple_New(argc);
+	PyObject *ret;
+
 	if (!pyargs)
 		return 0;
 
 	for (i = 0; i < argc; i++) {
-		PyTuple_SetItem(pyargs, i, PyString_FromString(argv[i]));
+		PyTuple_SetItem(pyargs, i, PyString_FromStringAndSize(argv[i], argvs[i]));
 	}
 
-	ret = python_call((PyObject *) func, pyargs, 0);
-
+	ret = python_call((PyObject *) func, pyargs, 0, NULL);
+	Py_DECREF(pyargs);
 	if (ret) {
 		if (PyString_Check(ret)) {
 			rv = PyString_AsString(ret);
-			rl = strlen(rv);
-			if (rl <= 0xffff) {
+			rl = PyString_Size(ret);
+			if (rl <= 65536) {
 				memcpy(buffer, rv, rl);
 				Py_DECREF(ret);
+				UWSGI_RELEASE_GIL;
 				return rl;
 			}
 		}
+		Py_DECREF(ret);
 	}
 
 	if (PyErr_Occurred())
 		PyErr_Print();
 
+	UWSGI_RELEASE_GIL;
 
 	return 0;
 
@@ -1143,8 +1354,9 @@ int uwsgi_python_spooler(char *filename, char *buf, uint16_t len, char *body, si
 		PyDict_SetItemString(spool_dict, "body", PyString_FromStringAndSize(body, body_len));
 	}
 	PyTuple_SetItem(pyargs, 0, spool_dict);
-	
-	ret = python_call(spool_func, pyargs, 0);
+
+	ret = python_call(spool_func, pyargs, 0, NULL);
+
 	if (ret) {
 		if (!PyInt_Check(ret)) {
 			// error, retry
@@ -1166,6 +1378,7 @@ int uwsgi_python_spooler(char *filename, char *buf, uint16_t len, char *body, si
 	return -1;
 }
 
+#ifndef UWSGI_PYPY
 void uwsgi_python_resume(struct wsgi_request *wsgi_req) {
 
 	PyThreadState *tstate = PyThreadState_GET();
@@ -1180,10 +1393,107 @@ void uwsgi_python_resume(struct wsgi_request *wsgi_req) {
 	}
 
 }
+#endif
 
 void uwsgi_python_fixup() {
 	// set hacky modifier 30
-	uwsgi.p[30] = uwsgi.p[0];
+	uwsgi.p[30] = uwsgi_malloc( sizeof(struct uwsgi_plugin) );
+	memcpy(uwsgi.p[30], uwsgi.p[0], sizeof(struct uwsgi_plugin) );
+	uwsgi.p[30]->init_thread = NULL;
+	uwsgi.p[30]->atexit = NULL;
+}
+
+void uwsgi_python_hijack(void) {
+
+	// the pyshell will be execute only in the first worker
+
+	FILE *pyfile;
+	if (up.pyrun) {
+		uwsgi.workers[uwsgi.mywid].hijacked = 1;
+		UWSGI_GET_GIL;
+		pyfile = fopen(up.pyrun, "r");
+		if (!pyfile) {
+			uwsgi_error_open(up.pyrun);
+			exit(1);
+		}
+		PyRun_SimpleFile(pyfile, up.pyrun);	
+		// could be never executed
+		exit(0);
+	}
+
+#ifndef UWSGI_PYPY
+	if (up.pyshell_oneshot && uwsgi.workers[uwsgi.mywid].hijacked_count > 0) {
+		uwsgi.workers[uwsgi.mywid].hijacked = 0;
+		return;
+	}
+	if (up.pyshell && uwsgi.mywid == 1) {
+		uwsgi.workers[uwsgi.mywid].hijacked = 1;
+		uwsgi.workers[uwsgi.mywid].hijacked_count++;
+		// re-map stdin to stdout and stderr if we are logging to a file
+		if (uwsgi.logfile) {
+			if (dup2(0, 1) < 0) {
+				uwsgi_error("dup2()");
+			}
+			if (dup2(0, 2) < 0) {
+				uwsgi_error("dup2()");
+			}
+		}
+		UWSGI_GET_GIL;
+		PyImport_ImportModule("readline");
+		int ret = PyRun_InteractiveLoop(stdin, "uwsgi");
+
+		if (up.pyshell_oneshot) {
+			exit(UWSGI_DE_HIJACKED_CODE);
+		}
+
+		if (ret == 0) {
+			exit(UWSGI_QUIET_CODE);
+		}
+		exit(0);
+	}
+#endif
+}
+
+int uwsgi_python_mule(char *opt) {
+
+	if (uwsgi_endswith(opt, ".py")) {
+		UWSGI_GET_GIL;
+		if (uwsgi_pyimport_by_filename("__main__", opt) == NULL) {
+			return 0;
+		}
+		UWSGI_RELEASE_GIL;
+		return 1;
+	}
+	
+	return 0;
+	
+}
+
+int uwsgi_python_mule_msg(char *message, size_t len) {
+
+	UWSGI_GET_GIL;
+
+	PyObject *mule_msg_hook = PyDict_GetItemString(up.embedded_dict, "mule_msg_hook");
+        if (!mule_msg_hook) {
+                // ignore
+                UWSGI_RELEASE_GIL;
+                return 0;
+        }
+
+	PyObject *pyargs = PyTuple_New(1);
+        PyTuple_SetItem(pyargs, 0, PyString_FromStringAndSize(message, len));
+
+        PyObject *ret = python_call(mule_msg_hook, pyargs, 0, NULL);
+	Py_DECREF(pyargs);
+	if (ret) {
+		Py_DECREF(ret);
+	}
+
+	if (PyErr_Occurred())
+                PyErr_Print();
+
+	UWSGI_RELEASE_GIL;
+	return 1;
 }
 
 struct uwsgi_plugin python_plugin = {
@@ -1194,10 +1504,10 @@ struct uwsgi_plugin python_plugin = {
 	.init = uwsgi_python_init,
 	.post_fork = uwsgi_python_post_fork,
 	.options = uwsgi_python_options,
-	.manage_opt = uwsgi_python_manage_options,
-	.short_options = "w:O:H:j:",
 	.request = uwsgi_request_wsgi,
 	.after_request = uwsgi_after_request_wsgi,
+
+	.preinit_apps = uwsgi_python_preinit_apps,
 	.init_apps = uwsgi_python_init_apps,
 
 	.fixup = uwsgi_python_fixup,
@@ -1207,18 +1517,28 @@ struct uwsgi_plugin python_plugin = {
 
 	.enable_threads = uwsgi_python_enable_threads,
 	.init_thread = uwsgi_python_init_thread,
-	.manage_xml = uwsgi_python_xml,
 
 	.magic = uwsgi_python_magic,
 
+#ifndef UWSGI_PYPY
 	.suspend = uwsgi_python_suspend,
 	.resume = uwsgi_python_resume,
+#endif
+
+	.hijack_worker = uwsgi_python_hijack,
+	.spooler_init = uwsgi_python_spooler_init,
 
 	.signal_handler = uwsgi_python_signal_handler,
 	.rpc = uwsgi_python_rpc,
 
+	.mule = uwsgi_python_mule,
+	.mule_msg = uwsgi_python_mule_msg,
+
 	.spooler = uwsgi_python_spooler,
 
-	.help = uwsgi_python_help,
+	.atexit = uwsgi_python_atexit,
+
+	.code_string = uwsgi_python_code_string,
+
 
 };
