@@ -3,16 +3,84 @@
 
 extern struct uwsgi_server uwsgi;
 
-static void spooler_readdir(char *);
+static void spooler_readdir(struct uwsgi_spooler *, char *dir);
 #ifdef __linux__
-static void spooler_scandir(char *);
+static void spooler_scandir(struct uwsgi_spooler *, char *dir);
 #endif
-void spooler_manage_task(char *, char *);
+static void spooler_manage_task(struct uwsgi_spooler *, char *, char *);
 
-// fake function to allow waking the spooler
-void spooler_wakeup() {}
+// increment it whenever a signal is raised
+static uint64_t wakeup = 0;
 
-pid_t spooler_start() {
+// function to allow waking up the spooler if blocked in event_wait
+void spooler_wakeup() { wakeup++; }
+
+void uwsgi_opt_add_spooler(char *opt, char *directory, void *none) {
+
+	int i;
+
+        if (access(directory, R_OK | W_OK | X_OK)) {
+                uwsgi_error("[spooler directory] access()");
+                exit(1);
+        }
+
+	if (uwsgi.spooler_numproc > 0) {
+		for(i=0;i<uwsgi.spooler_numproc;i++) {
+			uwsgi_new_spooler(directory);
+		}
+	}
+	else {
+        	uwsgi_new_spooler(directory);
+	}
+
+}
+
+
+struct uwsgi_spooler *uwsgi_new_spooler(char *dir) {
+
+        struct uwsgi_spooler *uspool = uwsgi.spoolers;
+
+        if (!uspool) {
+                uwsgi.spoolers = uwsgi_malloc_shared(sizeof(struct uwsgi_spooler));
+                uspool = uwsgi.spoolers;
+        }
+        else {
+                while(uspool) {
+                        if (uspool->next == NULL) {
+                                uspool->next = uwsgi_malloc_shared(sizeof(struct uwsgi_spooler));
+                                uspool = uspool->next;
+                                break;
+                        }
+                        uspool = uspool->next;
+                }
+        }
+
+        if (!realpath(dir, uspool->dir)) {
+                uwsgi_error("[spooler] realpath()");
+                exit(1);
+        }
+
+        uspool->next = NULL;
+
+        return uspool;
+}
+
+
+struct uwsgi_spooler *uwsgi_get_spooler_by_name(char *name) {
+
+	struct uwsgi_spooler *uspool = uwsgi.spoolers;
+
+	while(uspool) {
+		if (!strcmp(uspool->dir, name)) {
+			return uspool;
+		}
+		uspool = uspool->next;
+	}
+	
+	return NULL;
+}
+
+pid_t spooler_start(struct uwsgi_spooler *uspool) {
 	
 	int i;
 
@@ -23,16 +91,17 @@ pid_t spooler_start() {
 	}
 	else if (pid == 0) {
 		// USR1 will be used to wake up the spooler
-		signal(SIGUSR1, spooler_wakeup);
+		uwsgi_unix_signal(SIGUSR1, spooler_wakeup);
 		uwsgi.mywid = -1;
 		uwsgi.mypid = getpid();
+		uspool->pid = uwsgi.mypid;
 		// avoid race conditions !!!
-		uwsgi.shared->spooler_pid = uwsgi.mypid;
+		uwsgi.i_am_a_spooler = uspool;
 
-		uwsgi_fixup_fds(0, 0);
+		uwsgi_fixup_fds(0, 0, NULL);
 		uwsgi_close_all_sockets();
 
-		for (i = 0; i < 0xFF; i++) {
+		for (i = 0; i < 256; i++) {
                 	if (uwsgi.p[i]->post_fork) {
                         	uwsgi.p[i]->post_fork();
                 	}
@@ -40,7 +109,7 @@ pid_t spooler_start() {
 
 		uwsgi.signal_socket = uwsgi.shared->spooler_signal_pipe[1];
 
-		for (i = 0; i < 0xFF; i++) {
+		for (i = 0; i < 256; i++) {
                 	if (uwsgi.p[i]->spooler_init) {
                         	uwsgi.p[i]->spooler_init();
                 	}
@@ -52,10 +121,10 @@ pid_t spooler_start() {
                 	}
         	}
 
-		spooler();
+		spooler(uspool);
 	}
 	else if (pid > 0) {
-		uwsgi_log("spawned the uWSGI spooler on dir %s with pid %d\n", uwsgi.spool_dir, pid);
+		uwsgi_log("spawned the uWSGI spooler on dir %s with pid %d\n", uspool->dir, pid);
 	}
 
 	return pid;
@@ -78,32 +147,37 @@ void destroy_spool(char *dir, char *file) {
 }
 
 
-int spool_request(char *filename, int rn, int core_id, char *buffer, int size, char *priority, time_t at, char *body, size_t body_len) {
+int spool_request(struct uwsgi_spooler *uspool, char *filename, int rn, int core_id, char *buffer, int size, char *priority, time_t at, char *body, size_t body_len) {
 
 	struct timeval tv;
 	int fd;
 	struct uwsgi_header uh;
 
-	uwsgi_lock(uwsgi.spooler_lock);
+	if (!uspool) {
+		uspool = uwsgi.spoolers;
+	}
+
+	// this lock is for threads, the pid value in filename will avoid multiprocess races
+	uwsgi_lock(uspool->lock);
 
 	gettimeofday(&tv, NULL);
 
 	if (priority) {
-		if (snprintf(filename, 1024, "%s/%s", uwsgi.spool_dir, priority) <= 0) {
-			uwsgi_unlock(uwsgi.spooler_lock);
+		if (snprintf(filename, 1024, "%s/%s", uspool->dir, priority) <= 0) {
+			uwsgi_unlock(uspool->lock);
 			return 0;
 		}
 		// no need to check for errors...
 		(void) mkdir(filename, 0777);
 
-		if (snprintf(filename, 1024, "%s/%s/uwsgi_spoolfile_on_%s_%d_%d_%d_%llu_%llu", uwsgi.spool_dir, priority, uwsgi.hostname, (int) getpid(), rn, core_id, (unsigned long long) tv.tv_sec, (unsigned long long) tv.tv_usec) <= 0) {
-			uwsgi_unlock(uwsgi.spooler_lock);
+		if (snprintf(filename, 1024, "%s/%s/uwsgi_spoolfile_on_%s_%d_%d_%d_%llu_%llu", uspool->dir, priority, uwsgi.hostname, (int) getpid(), rn, core_id, (unsigned long long) tv.tv_sec, (unsigned long long) tv.tv_usec) <= 0) {
+			uwsgi_unlock(uspool->lock);
 			return 0;
 		}
 	}
 	else {
-		if (snprintf(filename, 1024, "%s/uwsgi_spoolfile_on_%s_%d_%d_%d_%llu_%llu", uwsgi.spool_dir, uwsgi.hostname, (int) getpid(), rn, core_id, (unsigned long long) tv.tv_sec, (unsigned long long) tv.tv_usec) <= 0) {
-			uwsgi_unlock(uwsgi.spooler_lock);
+		if (snprintf(filename, 1024, "%s/uwsgi_spoolfile_on_%s_%d_%d_%d_%llu_%llu", uspool->dir, uwsgi.hostname, (int) getpid(), rn, core_id, (unsigned long long) tv.tv_sec, (unsigned long long) tv.tv_usec) <= 0) {
+			uwsgi_unlock(uspool->lock);
 			return 0;
 		}
 	}
@@ -111,19 +185,16 @@ int spool_request(char *filename, int rn, int core_id, char *buffer, int size, c
 	fd = open(filename, O_CREAT | O_EXCL | O_WRONLY, S_IRUSR | S_IWUSR);
 	if (fd < 0) {
 		uwsgi_error_open(filename);
-		uwsgi_unlock(uwsgi.spooler_lock);
+		uwsgi_unlock(uspool->lock);
 		return 0;
 	}
 
-#ifdef __sun__
-	if (lockf(fd, F_LOCK, 0)) {
-		uwsgi_error("lockf()");
-#else
-	if (flock(fd, LOCK_EX)) {
-		uwsgi_error("flock()");
-#endif
+	// now lock the file, it will no be runnable, until the lock is not removed
+	// a race could come if the spooler take the file before fcntl is called
+        // in such case the spooler will detect a zeroed file and will retry later
+	if (uwsgi_fcntl_lock(fd)) {
 		close(fd);
-		uwsgi_unlock(uwsgi.spooler_lock);
+		uwsgi_unlock(uspool->lock);
 		return 0;
 	}
 
@@ -163,33 +234,47 @@ int spool_request(char *filename, int rn, int core_id, char *buffer, int size, c
 		}
 	}
 
+	// here the file will be unlocked too
 	close(fd);
 
-	uwsgi_log("[spooler] written %d bytes to file %s\n", size + body_len + 4, filename);
+	if (!uwsgi.spooler_quiet) 
+		uwsgi_log("[spooler] written %d bytes to file %s\n", size + body_len + 4, filename);
 	
-	uwsgi_unlock(uwsgi.spooler_lock);
+	// and here waiting threads can continue
+	uwsgi_unlock(uspool->lock);
 
-/*	wake up the spooler ... (HACKY) */
-	if (uwsgi.shared->spooler_pid > 0 ) {
-		(void) kill(uwsgi.shared->spooler_pid, SIGUSR1);
+/*	wake up the spoolers attached to the specified dir ... (HACKY) 
+	no need to fear races, as USR1 is harmless an all of the uWSGI processes...
+	it could be a problem if a new process takes the old pid, but modern systems should avoid that
+*/
+
+	struct uwsgi_spooler *spoolers = uwsgi.spoolers;
+	while(spoolers) {
+		if (!strcmp(spoolers->dir, uspool->dir)) {
+			if (spoolers->pid > 0 && spoolers->running == 0) {
+				(void) kill(spoolers->pid, SIGUSR1);
+			}
+		}
+		spoolers = spoolers->next;
 	}
 
 	return 1;
 
 
       clear:
-	uwsgi_unlock(uwsgi.spooler_lock);
+	uwsgi_unlock(uspool->lock);
 	uwsgi_error("write()");
 	if (unlink(filename)) {
 		uwsgi_error("unlink()");
 	}
+	// unlock the file too
 	close(fd);
 	return 0;
 }
 
 
 
-void spooler() {
+void spooler(struct uwsgi_spooler *uspool) {
 
 	// prevent process blindly reading stdin to make mess
 	int nullfd;
@@ -228,23 +313,28 @@ void spooler() {
 	for (;;) {
 
 
-		if (chdir(uwsgi.spool_dir)) {
+		if (chdir(uspool->dir)) {
 			uwsgi_error("chdir()");
 			exit(1);
 		}
 
 		if (uwsgi.spooler_ordered) {
 #ifdef __linux__
-			spooler_scandir(uwsgi.spool_dir);
+			spooler_scandir(uspool, NULL);
 #else
-			spooler_readdir(uwsgi.spool_dir);
+			spooler_readdir(uspool, NULL);
 #endif
 		}
 		else {
-			spooler_readdir(uwsgi.spool_dir);
+			spooler_readdir(uspool, NULL);
 		}
 
-		if (event_queue_wait(spooler_event_queue, uwsgi.shared->spooler_frequency, &interesting_fd) > 0) {
+		int timeout = uwsgi.shared->spooler_frequency;
+		if (wakeup > 0) {
+			timeout = 0;
+		}
+
+		if (event_queue_wait(spooler_event_queue, timeout, &interesting_fd) > 0) {
 			if (uwsgi.master_process) {
 				if (interesting_fd == uwsgi.shared->spooler_signal_pipe[1]) {
 					uint8_t uwsgi_signal;
@@ -264,15 +354,24 @@ void spooler() {
 			}
 		}
 
+		// avoid races
+		uint64_t tmp_wakeup = wakeup;
+		if (tmp_wakeup > 0) {
+			tmp_wakeup--;
+		}
+		wakeup = tmp_wakeup;
+
 	}
 }
 
 #ifdef __linux__
-static void spooler_scandir(char *dir) {
+static void spooler_scandir(struct uwsgi_spooler *uspool, char *dir) {
 
 	struct dirent **tasklist;
 	int n;
 
+	if (!dir) dir = uspool->dir;
+
 	n = scandir(dir, &tasklist, 0, versionsort);
 	if (n < 0) {
 		uwsgi_error("scandir()");
@@ -280,7 +379,7 @@ static void spooler_scandir(char *dir) {
 	}
 	
 	while(n--) {
-		spooler_manage_task(dir, tasklist[n]->d_name);
+		spooler_manage_task(uspool, dir, tasklist[n]->d_name);
 		free(tasklist[n]);	
 	}
 
@@ -289,15 +388,17 @@ static void spooler_scandir(char *dir) {
 #endif
 
 
-static void spooler_readdir(char *dir) {
+static void spooler_readdir(struct uwsgi_spooler *uspool, char *dir) {
 
 	DIR *sdir;
 	struct dirent *dp;
 
+	if (!dir) dir = uspool->dir;
+
 	sdir = opendir(dir);
 	if (sdir) {
 		while ((dp = readdir(sdir)) != NULL) {
-			spooler_manage_task(dir, dp->d_name);
+			spooler_manage_task(uspool, dir, dp->d_name);
 		}
 		closedir(sdir);
 	}
@@ -306,7 +407,7 @@ static void spooler_readdir(char *dir) {
 	}
 }
 
-void spooler_manage_task(char *dir, char *task) {
+void spooler_manage_task(struct uwsgi_spooler *uspool, char *dir, char *task) {
 
 	int i, ret;
 
@@ -317,6 +418,8 @@ void spooler_manage_task(char *dir, char *task) {
 
 	int spool_fd;
 
+	if (!dir) dir = uspool->dir;
+
 	if (!strncmp("uwsgi_spoolfile_on_", task, 19) || (uwsgi.spooler_ordered && is_a_number(task))) {
 		struct stat sf_lstat;
 		if (lstat(task, &sf_lstat)) {
@@ -335,7 +438,7 @@ void spooler_manage_task(char *dir, char *task) {
 				return;
 			}
 			char *prio_path = realpath(".", NULL);
-			spooler_scandir(prio_path);
+			spooler_scandir(uspool, prio_path);
 			free(prio_path);
 			if (chdir(dir)) {
 				uwsgi_error("chdir()");
@@ -347,33 +450,34 @@ void spooler_manage_task(char *dir, char *task) {
 			return;
 		}
 		if (!access(task, R_OK | W_OK)) {
-			uwsgi_log("[spooler] managing request %s ...\n", task);
 
-#ifdef __sun__
-			// lockf needs write permission
 			spool_fd = open(task, O_RDWR);
-#else
-			spool_fd = open(task, O_RDONLY);
-#endif
+
 			if (spool_fd < 0) {
-				uwsgi_error_open(task);
+				if (errno != ENOENT)
+					uwsgi_error_open(task);
 				return;
 			}
 
-#ifdef __sun__
-			if (lockf(spool_fd, F_LOCK, 0)) {
-				uwsgi_error("lockf()");
-#else
-			if (flock(spool_fd, LOCK_EX)) {
-				uwsgi_error("flock()");
-#endif
-				close(spool_fd);
+			// check if the file is locked by another process
+			if (uwsgi_fcntl_is_locked(spool_fd)) {
+				uwsgi_protected_close(spool_fd);
 				return;
 			}
 
-			if (read(spool_fd, &uh, 4) != 4) {
-				uwsgi_error("read()");
-				close(spool_fd);
+			// unlink() can destroy the lock !!!
+			if (access(task, R_OK | W_OK)) {
+				uwsgi_protected_close(spool_fd);
+				return;
+			}
+
+			ssize_t rlen = uwsgi_protected_read(spool_fd, &uh, 4);
+
+			if (rlen != 4) {
+				// it could be here for broken file or just opened one
+				if (rlen < 0)
+					uwsgi_error("read()");
+				uwsgi_protected_close(spool_fd);
 				return;
 			}
 
@@ -381,10 +485,10 @@ void spooler_manage_task(char *dir, char *task) {
 			uh.pktsize = uwsgi_swap16(uh.pktsize);
 #endif
 
-			if (read(spool_fd, spool_buf, uh.pktsize) != uh.pktsize) {
+			if (uwsgi_protected_read(spool_fd, spool_buf, uh.pktsize) != uh.pktsize) {
 				uwsgi_error("read()");
 				destroy_spool(dir, task);	
-				close(spool_fd);
+				uwsgi_protected_close(spool_fd);
 				return;
 			}			
 					
@@ -392,17 +496,23 @@ void spooler_manage_task(char *dir, char *task) {
 			if (sf_lstat.st_size > (uh.pktsize+4)) {
 				body_len = sf_lstat.st_size - (uh.pktsize+4);
 				body = uwsgi_malloc(body_len);
-				if ((size_t)read(spool_fd, body, body_len) != body_len) {
+				if ((size_t)uwsgi_protected_read(spool_fd, body, body_len) != body_len) {
 					uwsgi_error("read()");
 					destroy_spool(dir, task);
-					close(spool_fd);
+					uwsgi_protected_close(spool_fd);
 					free(body);
 					return;
 				}
 			}
 
-			close(spool_fd);
+			// now the task is running and should not be waken up
+			uspool->running = 1;
 
+			if (!uwsgi.spooler_quiet) 
+				uwsgi_log("[spooler %s pid: %d] managing request %s ...\n", uspool->dir, (int) uwsgi.mypid, task); 
+
+
+			// chdir before running the task (if requested)
 			if (uwsgi.spooler_chdir) {
 				if (chdir(uwsgi.spooler_chdir)) {
 					uwsgi_error("chdir()");
@@ -410,7 +520,7 @@ void spooler_manage_task(char *dir, char *task) {
 			}
 
 			int callable_found = 0;
-			for(i=0;i<0xff;i++) {
+			for(i=0;i<256;i++) {
 				if (uwsgi.p[i]->spooler) {
 					time_t now = time(NULL);
 					if(uwsgi.shared->options[UWSGI_OPTION_SPOOLER_HARAKIRI] > 0) {
@@ -423,7 +533,8 @@ void spooler_manage_task(char *dir, char *task) {
 					if (ret == 0) continue;
 					callable_found = 1;
 					if (ret == -2) {
-						uwsgi_log("[spooler] done with task %s after %d seconds\n", task, time(NULL)-now);
+						if (!uwsgi.spooler_quiet)
+							uwsgi_log("[spooler %s pid: %d] done with task %s after %d seconds\n", uspool->dir, (int) uwsgi.mypid, task, time(NULL)-now);
 						destroy_spool(dir, task);	
 					}
 					// re-spool it
@@ -431,9 +542,13 @@ void spooler_manage_task(char *dir, char *task) {
 				}
 			}
 
-			if (body) {
+			if (body)
 				free(body);
-			}
+
+			// here we free and unlock the task
+			uwsgi_protected_close(spool_fd);
+			uspool->running = 0;
+
 			if (!callable_found) {
 				uwsgi_log("unable to find the spooler function, have you loaded it into the spooler process ?\n");
 			}
@@ -442,46 +557,6 @@ void spooler_manage_task(char *dir, char *task) {
 	}
 }
 
-
-int uwsgi_request_spooler(struct wsgi_request *wsgi_req) {
-
-	int i;
-	char spool_filename[1024];
-
-	if (uwsgi.spool_dir == NULL) {
-		uwsgi_log("the spooler is inactive !!!...skip\n");
-		uwsgi_send_empty_pkt(wsgi_req->poll.fd, NULL, 255, 0);
-		return -1;
-	}
-
-	i = spool_request(spool_filename, uwsgi.workers[0].requests + 1, wsgi_req->async_id, wsgi_req->buffer, wsgi_req->uh.pktsize, NULL, 0, NULL, 0);
-	wsgi_req->uh.modifier1 = 255;
-	wsgi_req->uh.pktsize = 0;
-	if (i > 0) {
-		wsgi_req->uh.modifier2 = 1;
-		if (write(wsgi_req->poll.fd, wsgi_req, 4) != 4) {
-			uwsgi_log("disconnected client, remove spool file.\n");
-			/* client disconnect, remove spool file */
-			if (unlink(spool_filename)) {
-				uwsgi_error("unlink()");
-				uwsgi_log("something horrible happened !!! check your spooler ASAP !!!\n");
-				exit(1);
-			}
-		}
-		return 0;
-	}
-	else {
-		/* announce a failed spool request */
-		wsgi_req->uh.modifier2 = 0;
-		i = write(wsgi_req->poll.fd, wsgi_req, 4);
-		if (i != 4) {
-			uwsgi_error("write()");
-		}
-	}
-
-	return -1;
-}
-
 #else
 #warning "*** Spooler support is disabled ***"
 #endif
